import os, sys
import shutil


def pyonir_setup():
    from pyonir import PYONIR_SETUPS_DIRPATH
    from pyonir.utilities import copy_assets, PrntColrs

    base_path = os.getcwd()
    backend_dirpath = os.path.join(PYONIR_SETUPS_DIRPATH, 'backend')
    contents_dirpath = os.path.join(PYONIR_SETUPS_DIRPATH, 'contents')
    contents_slim_dirpath = os.path.join(PYONIR_SETUPS_DIRPATH, 'contents-slim')
    frontend_dirpath = os.path.join(PYONIR_SETUPS_DIRPATH, 'frontend')
    entry_filepath = os.path.join(PYONIR_SETUPS_DIRPATH, 'main.py')

    project_name = input(f"Whats your project name?").strip()
    project_path = os.path.join(base_path, project_name)
    use_demo = input(f"{PrntColrs.OKBLUE}Do you want to install the demo project?(y for yes, n for no){PrntColrs.RESET}").strip()
    if not os.path.exists(project_path):
        os.makedirs(project_path)
    use_frontend = input(f"{PrntColrs.OKBLUE}Do you need a frontend? (y for yes, n for no){PrntColrs.RESET}").strip()

    if use_demo.lower() == 'y':
        copy_assets(entry_filepath, os.path.join(project_path, 'main.py'), False)
        copy_assets(contents_dirpath, os.path.join(project_path, 'contents'), False)
        copy_assets(backend_dirpath, os.path.join(project_path, 'backend'), False)

        if use_frontend == 'y':
            copy_assets(frontend_dirpath, os.path.join(project_path, 'frontend'), False)

    summary = f'''{PrntColrs.OKGREEN}
Project {project_name} created!
- path: {project_path}
- use frontend: {use_frontend}{PrntColrs.RESET}
        '''
    print(summary)

def pyonir_install():
    import requests, zipfile, io
    """Installs plugins into pyonir application from the pyonir registry"""
    GIT_HUB_ZIP = "https://github.com/{repo_path}/archive/refs/heads/{repo_branch}.zip"
    project_base_dir = os.getcwd()
    action, context = sys.argv[1:]
    if action == 'install':
        if context.startswith('plugins:'):
            dir_name, repo_context = context.split(':')
            repo_path, repo_branch = repo_context.split('#')
            _, repo_name = repo_path.split('/')
            repo_zip = GIT_HUB_ZIP.format(repo_path=repo_path, repo_branch=repo_branch)
            temp_dst_path = os.path.join(project_base_dir, dir_name, "."+repo_name)
            dst_path = os.path.join(project_base_dir, dir_name, repo_name)
            print(f"pyonir is downloading {repo_zip} ...")
            response = requests.get(repo_zip)
            response.raise_for_status()
            if not os.path.exists(temp_dst_path):
                os.makedirs(temp_dst_path)
            with zipfile.ZipFile(io.BytesIO(response.content)) as zf:
                zf.extractall(temp_dst_path)
            extracted_folder = os.path.join(temp_dst_path, f"{repo_name}-{repo_branch}")
            shutil.move(extracted_folder, dst_path)
            shutil.rmtree(temp_dst_path)

    print(action, context)
    pass

if __name__ == '__main__':
    pyonir_install()


from dataclasses import dataclass, field
# from enum import StrEnum
import os
from enum import Enum
from typing import Optional, Union, Callable, List, Tuple, Iterator

from starlette.applications import Starlette

TEXT_RES: str = 'text/html'
JSON_RES: str = 'application/json'
EVENT_RES: str = 'text/event-stream'
PAGINATE_LIMIT: int = 6


# === Route Definitions ===

RoutePath = str
"""Represents the URL path of a route (e.g., '/about', '/api/data')."""

RouteFunction = Callable
"""A callable that handles a specific route request (e.g., controller function)."""

RouteMethods = List[str]
"""HTTP methods supported by a route (e.g., ['GET', 'POST'])."""

PyonirRoute = Tuple[RoutePath, RouteFunction, RouteMethods]
"""A single route entry containing the path, its handler function, and allowed HTTP methods."""

PyonirRouters = List[Tuple[RoutePath, List[PyonirRoute]]]
"""A list of route groups, each keyed by a base path and containing multiple PyonirRoute entries."""


# === Application Module Definitions ===

AppName = str
"""The name identifier for an app module."""

ModuleName = str
"""The Python module name used for import or registration."""

AppEndpoint = str
"""The base endpoint path where the app is mounted."""

AppPaths = List[str]
"""A list of file or URL paths associated with the app."""

AppContentsPath = str
"""The root path to the static or content files of the app."""

AppSSGPath = str
"""The path used for static site generation output."""

AppContextPaths = Tuple[AppName, RoutePath, AppPaths]
"""Context binding tuple that connects an app name to a route and its associated paths."""

AppCtx = Tuple[ModuleName, RoutePath, AppContentsPath, AppSSGPath]
"""Full application context including module reference and content/static paths."""

AppRequestPaths = Tuple[RoutePath, AppPaths]
"""Tuple representing an incoming request path and all known paths for resolution."""

@dataclass
class ParselyPagination:
    limit: int = 0
    max_count: int = 0
    curr_page: int = 0
    page_nums: list[int, int] = field(default_factory=list)
    items: list['Parsely'] = field(default_factory=list)

    def __iter__(self) -> Iterator['Parsely']:
        return iter(self.items)


@dataclass
class Theme:
    _mapper = {'theme_dirname': 'file_dirname', 'theme_dirpath': 'file_dirpath'}
    name: str
    theme_dirname: str = ''
    theme_dirpath: str = ''
    static_dirname: str = 'static' # path to serve theme's static assets (css,js,images) used to style the UI
    templates_dirname: str = 'layouts' # path to serve theme's template files used to rendering HTML pages

    @property
    def static_dirpath(self):
        """directory to serve static theme assets"""
        return os.path.join(self.theme_dirpath, self.static_dirname)

    @property
    def jinja_template_path(self):
        return os.path.join(self.theme_dirpath, self.templates_dirname)

class PyonirOptions:
    contents_dirpath: str  = '' # base directory path for markdown files
    use_file_based_routing: bool = None # toggles use of file based routing directory
    routes_dirpath: str = ''    # path for resolving file based routing
    routes_api_dirpath: str = '' # path for resolving file based API configured endpoints
    file_based_routes: dict = {}# configurations for resolving file based routes that contain dynamic path params

class Pagination:
    page_num: int = 1
    limit: int = 0

class PyonirHooks(str):
    AFTER_INIT = 'AFTER_INIT'
    ON_REQUEST = 'ON_REQUEST'
    ON_PARSELY_COMPLETE = 'ON_PARSELY_COMPLETE'

class PyonirServer(Starlette):
    ws_routes = []
    sse_routes = []
    auth_routes = []
    endpoints = []
    url_map = {}
    resolvers = {}
    services = {}
    paginate: Pagination = Pagination()

    def response_renderer(self): pass
    def serve_redirect(self): pass
    def create_endpoint(self): pass
    def create_route(self): pass
    def serve_static(self): pass


class Parsely:
    default_file_attributes = ['file_name','file_path','file_dirname','file_data_type','file_ctx','file_created_on']
    resolver: Optional[callable]
    route: Optional[callable]
    app_ctx: AppCtx # application context for file
    file_status: str
    file_path: str
    file_dirpath: str # path to files contents directory
    file_contents: str # contents of a file
    file_contents_dirpath: str # contents directory path used when querying refs
    data: dict
    schema: any
    is_page: bool
    is_home: bool # is home page of site
    file_ctx: str # the application context name
    file_dirname: str # nearest parent directory for file
    file_data_type: str # data type based on root contents directory name
    file_name: str # the file name
    file_ext: str # the file extenstion
    file_exists: bool # determines if file exists on filesystem
    file_ssg_api_dirpath: str # the files static generated api endpoint
    file_ssg_html_dirpath: str # the files static generated html endpoint

    async def process_route(self, pyonir_request, app_ctx): pass

    async def process_resolver(self, pyonir_request): pass

    async def process_response(self, pyonir_request): pass


class PyonirCollection:
    SortedList: list
    get_attr: callable
    dict_to_class: callable
    collections: list[Parsely]

class PyonirThemes:
    """Represents sites available and active theme(s) within the frontend directory."""
    themes_dirpath: str # directory path to available site themes
    _available_themes: PyonirCollection | None # collection of themes available in frontend/themes directory

    @property
    def active_theme(self) -> Theme | None: pass

    def _get_available_themes(self) -> PyonirCollection | None: pass

class TemplateEnvironment:
    themes: PyonirThemes

class PyonirRequest:

    server_response: any
    file: Optional[Parsely]
    server_request: Optional['StarletteRequest']
    raw_path: str
    method: str
    path: str
    url: str
    slug: str
    path_params: dict[str, str]
    query_params: dict[str, str]
    parts: list[str]
    limit: int
    model: str
    is_home: bool
    is_api: bool
    form: dict
    files: list
    ip: str
    host: str
    protocol: str
    headers: dict
    browser :str
    type: Union[TEXT_RES, JSON_RES, EVENT_RES]
    status_code: int

    @property
    def redirect(self): pass

    async def process_request_data(self):
        """Get form data and file upload contents from request"""
        pass

    def derive_status_code(self, is_router_method: bool):
        """Create status code for web request based on a file's availability, status_code property"""
        pass

    def render_error(self):
        """Data output for an unknown file path for a web request"""
        pass

    def resolve_request_to_file(self, path_str: str, app: 'PyonirApp', skip_vanity: bool = False) -> tuple['PyonirApp', Parsely]:pass

    @staticmethod
    def process_header(headers):pass

    @staticmethod
    def get_params(url):pass


class PyonirBase:
    PAGINATE_LIMIT: str
    DATE_FORMAT: str
    TIMEZONE: str
    SOFTWARE_VERSION: str
    APPS_DIRNAME: str
    BACKEND_DIRNAME: str
    FRONTEND_DIRNAME: str
    CONTENTS_DIRNAME: str
    THEMES_DIRNAME: str
    CONFIGS_DIRNAME: str
    TEMPLATES_DIRNAME: str
    SSG_DIRNAME: str
    SSG_IN_PROGRESS: str
    UPLOADS_THUMBNAIL_DIRNAME: str
    UPLOADS_DIRNAME: str
    ASSETS_DIRNAME: str
    API_DIRNAME: str
    PAGES_DIRNAME: str
    CONFIG_FILENAME: str

class PyonirApp(PyonirBase):
    configs: object
    request_paths: str
    nginx_config_filepath: str
    unix_socket_filepath: str
    ssg_dirpath: str
    logs_dirpath: str
    backend_dirpath: str
    contents_dirpath: str
    frontend_dirpath: str
    pages_dirpath: str
    api_dirpath: str
    plugins_dirpath: str
    uploads_dirpath: str
    resolvers_dirpath: str
    jinja_filters_dirpath: str
    app_ctx: str
    env: str
    is_dev: str
    host: str
    port: str
    protocol: str
    is_secure: str
    domain: str
    parse_jinja: callable
    parse_pyformat: callable
    setup_templates: callable
    install_sys_plugins: callable
    install_plugins: callable
    run_plugins: callable
    run_async_plugins: callable
    setup_configs: callable
    run: callable
    pyonir_path: type
    endpoint: str
    app_dirpath: str
    app_name: str
    app_account_name: str
    TemplateEnvironment: TemplateEnvironment
    server: PyonirServer
    available_plugins: dict[str, callable]

class PyonirPlugin(PyonirBase):

    _app_ctx = None
    app: PyonirApp
    app_entrypoint: str # plugin application initializing file
    app_dirpath: str # plugin directory path
    name: str # web url to serve application pages
    routing_paths: list
    CONFIG_FILENAME: str
    request_paths: AppRequestPaths
    backend_dirpath: str
    contents_dirpath: str
    frontend_dirpath: str
    ssg_dirpath: str
    app_ctx: AppCtx

    def register_routing_dirpaths(self, dir_paths: list):
        """Registers a new pages directory path for resolving web based request"""
        pass

    def register_templates(self, dir_paths: list):
        """Registers additional paths for jinja templates. Templates will load in order of priority."""
        pass

    def insert(self, file_path: str, contents: dict) -> Parsely:
        """Creates a new file"""
        pass

    @staticmethod
    def query_files(dir_path: str, app_ctx: tuple, model_type: any = None) -> list[Parsely]: pass

    @staticmethod
    def install_directory(plugin_src_directory: str, site_destination_directory: str): pass


import asyncio
import os, typing, json, inspect
from typing import get_type_hints

from starlette.websockets import WebSocket, WebSocketState, WebSocketDisconnect

# from pyonir import ASSETS_ROUTE, UPLOADS_ROUTE, PAGINATE_LIMIT, API_ROUTE, API_DIRNAME
from pyonir.pyonir_types import PyonirApp, PyonirRequest, PyonirServer, PyonirHooks
from pyonir.utilities import create_file, get_attr, cls_mapper

TEXT_RES = 'text/html'
JSON_RES = 'application/json'
EVENT_RES = 'text/event-stream'

ConnClients = {}


async def pyonir_ws_handler(websocket: WebSocket):
    """ws connection handler"""
    from pyonir.utilities import generate_id

    async def get_data(ws: WebSocket):
        assert ws.application_state == WebSocketState.CONNECTED and ws.client_state == WebSocketState.CONNECTED
        wsdata = await ws.receive()

        if wsdata.get('text'):
            wsdata = wsdata['text']
            swsdata = json.loads(wsdata)
            swsdata['value'] = swsdata.get('value')
            wsdata = json.dumps(swsdata)
        elif wsdata.get('bytes'):
            wsdata = wsdata['bytes'].decode('utf-8')

        return wsdata

    async def broadcast(message: str, ws_id: str = None):
        for id, ws in ConnClients.items():
            if active_id == id and hasattr(ws, 'send_text'): continue
            await ws.send_text(message)

    async def on_disconnect(websocket: WebSocket):
        del ConnClients[active_id]
        client_data.update({"action": "ON_DISCONNECTED", "id": active_id})
        await broadcast(json.dumps(client_data))

    async def on_connect(websocket: WebSocket):
        client_data.update({"action": "ON_CONNECTED", "id": active_id})
        await websocket.send_text(json.dumps(client_data))

    active_id = generate_id()
    client_data = {}
    await websocket.accept()  # Accept the WebSocket connection
    print("WebSocket connection established!")
    ConnClients[active_id] = websocket
    await on_connect(websocket)
    try:
        while websocket.client_state == WebSocketState.CONNECTED:
            # Wait for a message from the client
            data = await get_data(websocket)
            print(f"Received from client: {data}")
            # Respond to the client
            await broadcast(data)
        await on_disconnect(data)
    except Exception as e:
        del ConnClients[active_id]
        print(f"WebSocket connection closed: {active_id}")

async def pyonir_sse_handler(request: PyonirRequest) -> typing.AsyncGenerator:
    """Handles sse web request by pyonir"""
    from pyonir.utilities import generate_id
    request.type = EVENT_RES  # assign the appropriate streaming headers
    # set sse client
    event = get_attr(request.query_params, 'event')
    retry = get_attr(request.query_params, 'retry') or 1000
    close_id = get_attr(request.query_params, 'close')
    interval = 1  # time between events
    client_id = get_attr(request.query_params, 'id') or request.headers.get('user-agent')
    client_id += f"{client_id}{generate_id()}"
    if close_id and ConnClients.get(close_id):
        del ConnClients[close_id]
        return
    last_client = ConnClients.get(client_id, {
        "retry": retry,
        "event": event,
        "id": client_id,
        "data": {
            "time": 0
        },
    })
    # register new client
    if not ConnClients.get(client_id):
        ConnClients[client_id] = last_client

    while True:
        last_client["data"]["time"] = last_client["data"]["time"] + 1
        is_disconnected = await request.server_request.is_disconnected()
        if is_disconnected or close_id:
            del ConnClients[client_id]
            break
        await asyncio.sleep(interval)  # Wait for 5 seconds before sending the next message
        res = process_sse(last_client)
        yield res

async def pyonir_docs_handler(request: PyonirRequest):
    """Documentation for every endpoint by pyonir"""
    return request.server_request.app.url_map

def pyonir_index(request: PyonirRequest):
    """Catch all routes for all web request"""
    pass


def setup_starlette_server(iapp: PyonirApp) -> PyonirServer:
    """Setup Starlette web server"""

    from starlette_wtf import CSRFProtectMiddleware
    # from starlette_session import SessionMiddleware
    from starlette.middleware.sessions import SessionMiddleware

    from starlette.middleware.trustedhost import TrustedHostMiddleware
    from starlette.middleware.gzip import GZipMiddleware
    from starlette.responses import Response, StreamingResponse
    # from starlette.requests import Request
    from starlette.routing import Router, Route

    def render_response(value, media_type):
        if media_type == EVENT_RES:
            return StreamingResponse(content=value, media_type=media_type)
        return Response(content=value, media_type=media_type)

    def redirect(url, code=302):
        from starlette.responses import RedirectResponse
        res = RedirectResponse(url=url.strip(), status_code=code)
        res.headers["Cache-Control"] = "no-store, no-cache, must-revalidate, max-age=0"
        res.headers["Pragma"] = "no-cache"
        res.headers["Expires"] = "0"
        return res

    def get_staticroute(assets_dir):
        from starlette.staticfiles import StaticFiles
        return StaticFiles(directory=assets_dir)

    allowed_hosts = ['localhost', '*.localhost']
    star_app = PyonirServer()  # inherits from Starlette
    star_app.add_middleware(TrustedHostMiddleware)
    # star_app.add_middleware(GZipMiddleware, minimum_size=500)
    star_app.add_middleware(SessionMiddleware,
                            https_only=False,
                            secret_key=iapp.SECRET_SAUCE,
                            session_cookie=iapp.SESSION_KEY,
                            same_site='lax'
                            )
    star_app.add_middleware(CSRFProtectMiddleware, csrf_secret=iapp.SECRET_SAUCE)

    # Interface properties required for pyonir to process web request
    setattr(star_app, 'response_renderer', render_response)
    setattr(star_app, 'serve_redirect', redirect)
    setattr(star_app, 'create_endpoint', Router)
    setattr(star_app, 'create_route', Route)
    setattr(star_app, 'serve_static', get_staticroute)

    return star_app


def url_for(name, attr='path'):
    """returns application route value based on named class function name"""
    from pyonir import Site
    if not Site: return None
    urlmap = Site.server.url_map
    return urlmap.get(name, {}).get(attr)


def init_app_endpoints(endpoints: 'Endpoints'):
    for endpoint, routes in endpoints:
        for path, func, methods, *opts in routes:
            _add_route(func, f'{endpoint}{path}', methods, *opts)
            pass


def init_pyonir_endpoints(app: PyonirApp):
    app_theme = app.TemplateEnvironment.themes.active_theme
    for r, static_abspath in ((app.ASSETS_ROUTE, app_theme.static_dirpath), (app.UPLOADS_ROUTE, app.uploads_dirpath)):
        if not os.path.exists(static_abspath): continue
        _add_route(None, r, static_path=static_abspath)

    _add_route(pyonir_ws_handler, "/sysws", ws=True)
    _add_route(pyonir_index, "/", methods='*')
    _add_route(pyonir_index, "/{path:path}", methods='*')

def process_sse(data: dict) -> str:
    """Formats a string and an event name in order to follow the event stream convention.
    'event: Jackson 5\\ndata: {"abc": 123}\\n\\n'
    """
    sse_payload = ""
    for key, val in data.items():
        val = json.dumps(val) if key == 'data' else val
        sse_payload += f"{key}: {val}\n"
    return sse_payload + "\n"


def serve_favicon(app: PyonirApp):
    from starlette.responses import FileResponse
    return FileResponse(os.path.join(app.TemplateEnvironment.themes.active_theme.static_dirpath,'favicon.ico'), 200)

def route(dec_func: typing.Callable | None,
               path: str = '',
               methods=None,
               models: dict = None,
               auth: bool = None,
               ws: bool = None,
               sse: bool = None,
               static_path: str = None) -> typing.Callable | None:
    if dec_func is None and static_path is None:
        dec_func = pyonir_index
    return _add_route(dec_func, path, methods, models, auth, ws, sse, static_path)

def _add_route(dec_func: typing.Callable | None,
               path: str = '',
               methods=None,
               models: dict = None,
               auth: bool = None,
               ws: bool = None,
               sse: bool = None,
               static_path: str = None) -> typing.Callable | None:
    """Route decorator"""
    from pyonir import Site

    is_async = inspect.iscoroutinefunction(dec_func) if dec_func else False
    is_asyncgen = inspect.isasyncgenfunction(dec_func) if dec_func else False
    # list_of_args = list(inspect.signature(dec_func).parameters.keys()) if dec_func else None
    if methods == '*':
        methods = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']
    if methods is None:
        methods = ['GET']

    if static_path:
        from starlette.staticfiles import StaticFiles
        static_route = Site.server.serve_static(static_path)
        Site.server.mount(path, static_route)
        Site.server.url_map[path.lstrip('/')
        .replace('/', '_')] = {'path': path, 'dir': static_path, 'exists': os.path.exists(static_path)}
        return

    route_name = dec_func.__name__
    docs = dec_func.__doc__
    route_path = path.split('/{')[0]
    name = route_name
    endpoint_route = path.split('/', 1)[0]
    default_system_router = dec_func.__name__ == 'pyonir_index'
    req_models = Site.server.url_map.get(route_name, {}).get('models') or {}
    if models:
        for req_param, req_model in models.items():
            req_models[req_param] = req_model.__name__
            if hasattr(Site.server, 'models'):
                Site.server.models.update({req_model.__name__, req_model})
            else:
                Site.server.models = {req_model.__name__, req_model}
    new_route = {
        "doc": docs,
        "endpoint": endpoint_route,
        "params": {k: v.__name__ for k, v in dec_func.__annotations__.items()},
        "route": path,  # has regex pattern
        "path": route_path,
        "methods": methods,
        "models": models or req_models,
        "name": name,
        "auth": auth,
        "sse": sse,
        "ws": ws,
        "async": is_async,
        "async_gen": is_asyncgen,
    }
    # Add route path into categories
    Site.server.endpoints.append(f"{endpoint_route}{route_path}")
    Site.server.url_map[name] = new_route
    if sse:
        Site.server.sse_routes.append(f"{endpoint_route}{route_path}")
    if auth:
        Site.server.auth_routes.append(f"{endpoint_route}{route_path}")
    if ws:
        Site.server.ws_routes.append(f"{endpoint_route}{route_path}")
        return Site.server.add_websocket_route(path, dec_func, dec_func.__name__)

    async def dec_wrapper(star_req):
        from pyonir.core import PyonirRequest
        if star_req.url.path == '/favicon.ico': return serve_favicon(Site)
        # Resolve page file route
        pyonir_request = PyonirRequest(star_req)
        await pyonir_request.process_request_data()
        # app_ctx, req_filepath = resolve_path_to_file(star_req.url.path, Site)

        # Update template global
        Site.TemplateEnvironment.globals['request'] = pyonir_request
        pyonir_request.is_api = pyonir_request.parts and pyonir_request.parts[0] == Site.API_DIRNAME

        # Query File system for Page from request
        app_ctx, req_file = pyonir_request.resolve_request_to_file(star_req.url.path, Site)
        pyonir_request.file = req_file
        pyonir_request.app_ctx_name = app_ctx.name

        # Preprocess routes or resolver endpoint from file
        await req_file.process_route(pyonir_request, app_ctx)
        await req_file.process_resolver(pyonir_request)
        route_func = dec_func if not req_file.resolver else req_file.resolver


        # Get router endpoint from map
        if callable(route_func):
            is_async = inspect.iscoroutinefunction(route_func)

            default_args = dict(typing.get_type_hints(route_func))
            default_args.update(**pyonir_request.path_params)
            default_args.update(**pyonir_request.query_params.__dict__)
            default_args.update(**pyonir_request.form)
            args = cls_mapper(default_args, route_func, from_request=pyonir_request)
            pyonir_request.router_args = args

            # Resolve route decorator methods
            pyonir_request.server_response = await route_func(**args) if is_async else route_func(**args)

        else:
            pyonir_request.server_response = req_file.resolver or req_file.route
        try:

            # Execute plugins hooks initial request
            pyonir_request.derive_status_code(Site.server.url_map.get(route_func.__name__) and not default_system_router)
            await Site.run_async_plugins(PyonirHooks.ON_REQUEST, pyonir_request)

            # Finalize response output
            if pyonir_request.path not in Site.server.sse_routes + Site.server.ws_routes:
                pyonir_request.server_response = await req_file.process_response(pyonir_request)

        except Exception as e:
            raise
            # pyonir_request.server_response = json.dumps({"error": f"{e}"})
        if pyonir_request.redirect:
            return Site.server.serve_redirect(pyonir_request.redirect)

        return build_response(pyonir_request)

    dec_wrapper.__name__ = dec_func.__name__
    dec_wrapper.__doc__ = dec_func.__doc__
    Site.server.add_route(path, dec_wrapper, methods=methods)


def build_response(request: PyonirRequest):
    """Create web response for web server"""
    from datetime import datetime, timedelta
    from pyonir import Site
    force_fresh = "no-store, no-cache, must-revalidate, max-age=0"
    force_cache = "public"

    # ishtml = request.type == TEXT_RES
    # expires = datetime.utcnow() + timedelta(days=7)
    # expires = expires.strftime("%a, %d %b %Y %H:%M:%S GMT")
    response = Site.server.response_renderer(request.server_response, media_type=request.type)
    # if ishtml: response.headers['Expires'] = expires
    response.headers['Cache-Control'] = force_fresh
    response.headers['Pragma'] = "no-cache"
    response.headers['Expires'] = "0"
    # response.headers['Vary'] = "Cookie"
    response.headers['Server'] = "Pyonir Web Framework"
    response.status_code = request.status_code
    return response


def get_route_ctx(app: PyonirApp, path_str: str) -> tuple:
    """Gets the routing context from web request"""
    path_str = path_str.replace('api/', '')
    req_trgt = path_str.split('/').pop()
    res = None
    for ctxp in app.request_paths:
        app_ctx, route, pths, prunes = ctxp
        ishome = route == path_str
        prunes = [p for p in prunes if p != req_trgt or ishome]
        if not path_str.startswith(route.lstrip('/')): continue
        if ishome: route = ''
        res = (app if isinstance(app_ctx, str) else app_ctx), route, pths, prunes
    return res


def resolve_path_to_file(path_str: str, app: PyonirApp, skip_vanity: bool = False) -> typing.Tuple[PyonirApp, str]:
    from pyonir import Site
    path_result = None
    is_home = path_str == '/'
    if not is_home and hasattr(app, 'available_plugins'):
        for plg in app.available_plugins:
            if not hasattr(plg, 'request_paths') or not plg.endpoint: continue
            plg, path_result = resolve_path_to_file(path_str, plg)
            if path_result: break
        if path_result: return plg, path_result

    ctx_route, ctx_paths = app.request_paths
    check_wildcard = ctx_route!='/' and (path_str.startswith(ctx_route) or path_str.replace('/api','').startswith(ctx_route))
    # ctx_route = ctx_route.replace('/*','')
    ctx_slug = ctx_route[1:]
    path_slug = path_str[1:]
    reqst_path = [p for p in path_slug.split('/') if p not in ('api', ctx_slug)]
    if path_str.startswith('/api'): path_str = path_str.replace('/api','')
    if not is_home and not path_str.startswith(ctx_route): return app, None

    for rootp in ctx_paths:
        is_cat = os.path.join(rootp, *reqst_path, 'index.md')
        is_page = os.path.join(rootp, *reqst_path) + Site.EXTENSIONS['file']
        for pp in (is_cat, is_page):
            if not os.path.exists(pp): continue
            path_result = pp
            break
        if path_result: break
    if path_result is None and check_wildcard:
        dynamic_paths = [os.path.join(app.pages_dirpath, *reqst_path[:i],"index.md") for i in range(len(reqst_path), -1, -1) if os.path.exists(os.path.join(app.pages_dirpath, *reqst_path[:i],"index.md"))]
        path_result = dynamic_paths[0]
    return app, path_result


def generate_nginx_conf(app: PyonirApp):
    """Generates a NGINX conf file based on App configurations"""
    nginxconf = app.TemplateEnvironment.get_template("nginx.jinja.conf") \
        .render(
        app_name=app.configs.app.name,
        app_name_id=app.configs.app.name.replace(' ', '_').lower(),
        domain=app.domain,
        is_secure=app.is_secure,
        serve_static=True,
        site_dirpath=app.app_dirpath,
        site_logs_dirpath=app.logs_dirpath,
        app_socket_filepath=app.unix_socket_filepath,
        site_assets_route=app.ASSETS_ROUTE,
        site_theme_assets_dirpath=app.TemplateEnvironment.themes.active_theme.static_dirpath,
        site_uploads_route=app.UPLOADS_ROUTE,
        site_uploads_dirpath=app.uploads_dirpath,
        site_ssg_dirpath=app.ssg_dirpath,
        custom_nginx_locations=get_attr(app.configs, 'nginx_locations')
    )

    create_file(app.app_nginx_conf_filepath, nginxconf, False)

# def gather_file_based_routing(app: PyonirApp):
#     from pyonir.types import PyonirCollection, PyonirSchema
#     router_pages = PyonirCollection.query(app.pages_dirpath, app_ctx=app.app_ctx,
#                                           include_only="index.md")
#     router_pages = list(router_pages.where('@routes'))
#     pass

def start_uvicorn_server(app: PyonirApp, endpoints: 'Endpoints'):
    """Starts the webserver"""
    import uvicorn

    # """Uvicorn web server configurations"""
    # from pyonir import PYONIR_SSL_KEY, PYONIR_SSL_CRT
    uvicorn_options = {
        "port": app.port,
        "host": app.host
    }
    # if app.is_secure:
    #     uvicorn_options["ssl_keyfile"] = PYONIR_SSL_KEY
    #     uvicorn_options["ssl_certfile"] = PYONIR_SSL_CRT
    if not app.is_dev:
        uvicorn_options['uds'] = app.app_socket_filename

    # Initialize routers
    # gather_file_based_routing(app)
    init_app_endpoints(endpoints)
    init_pyonir_endpoints(app)
    print(f"/************** ASGI APP SERVER RUNNING on {'http' if app.is_dev else 'sock'} ****************/")
    print(f"\
    \n\t- App env: {'DEV' if app.is_dev else 'PROD'}\
    \n\t- App name: {app.name}\
    \n\t- App domain: {app.domain}\
    \n\t- App host: {app.host}\
    \n\t- App port: {app.port}\
    \n\t- App sock: {app.unix_socket_filepath}\
    \n\t- App Server: Uvicorn \
    \n\t- NGINX config: {app.nginx_config_filepath} \
    \n\t- System Version: {app.SOFTWARE_VERSION}")
    app.run_plugins(PyonirHooks.AFTER_INIT)
    uvicorn.run(app.server, **uvicorn_options)



import unittest, os
true = True
class ParselyTests(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        from pyonir.parser import Parsely
        from pyonir import init
        app = init(__file__)
        cls.parselyFile = Parsely(os.path.join(os.path.dirname(__file__),'contents','test.md'), app.app_ctx)

    def test_basic(self):
        self.assertEqual("scalar value", self.parselyFile.data.get('basic'))


    def test_dict_value(self):
        self.assertEqual({"my_key": "my_value", "another_key": "another_value"}, self.parselyFile.data.get('dict_value'))


    def test_list_value(self):
        self.assertEqual(["one", "two", "three"], self.parselyFile.data.get('list_value'))


    def test_dynamic_list_blocks(self):
        self.assertEqual([{"ages": [1, true, "hello", 3.14, {"dict_key": "dict_value"}]}, {"this": {"age": 3, "key": "some value"}}], self.parselyFile.data.get('dynamic_list_blocks'))


    def test_inline_list_of_scalrs_types(self):
        self.assertEqual([1, true, "hello", 3.14], self.parselyFile.data.get('inline_list_of_scalrs_types'))


    def test_inline_list_of_maps(self):
        self.assertEqual([{"one": 1}, {"two": true}, {"three": "hello"}], self.parselyFile.data.get('inline_list_of_maps'))


    def test_inline_dict_value(self):
        self.assertEqual({"my_lnkey": "my_lnvalue", "another_lnkey": "another_lnvalue"}, self.parselyFile.data.get('inline_dict_value'))


    def test_multiline_block(self):
        self.assertEqual("What is this here? Content types enable you to organize and manage content in a consistent way for specific kinds of pages.\nthere is no such thing as a Python JSON object. JSON is a language independent file \nformat that finds its roots in JavaScript, and is supported by many languages. end of mulitiline block.\n", self.parselyFile.data.get('multiline_block'))


    def test_content(self):
        self.assertEqual("What is this here? Content types enable you to organize and manage content in a consistent way for specific kinds of pages.\nthere is no such thing as a Python JSON object. JSON is a language independent file \nformat that finds its roots in JavaScript, and is supported by many languages. If your YAML\n", self.parselyFile.data.get('content'))




import dataclasses
import os

from pyonir.pyonir_types import PyonirRequest, PyonirApp
from pyonir.core import PyonirPlugin, PyonirCollection

@dataclasses.dataclass
class Menu:
    _mapper_key = 'menu'
    url: str
    slug: str = ''
    title: str = ''
    group: str = ''
    parent: str = ''
    icon: str = ''
    img: str = ''
    rank: int = 0
    subtitle: str = ''
    dom_class: str = ''
    status: str = ''

    def __post_init__(self):
        self.name = self.title
        pass

class Navigation(PyonirPlugin):
    """Assembles a map of navigation menus based on file configurations"""
    name = 'Navigation Plugin'

    def __init__(self, app: PyonirApp):
        self.menus = {}
        self.active_page = None
        self.build_navigation(app=app)
        # include navigation template example
        self.app = app
        self.register_templates([os.path.join(os.path.dirname(__file__), 'templates')])
        pass

    def after_init(self, data: any, app: PyonirApp):
        self.build_plugins_navigation(app)

    async def on_request(self, request: PyonirRequest, app: PyonirApp):
        """Executes task on web request"""
        refresh_nav = bool(getattr(request.query_params,'rnav', None))
        curr_nav = app.TemplateEnvironment.globals.get('navigation')
        if curr_nav and not refresh_nav: return None
        self.active_page = request.path
        self.build_navigation(app)
        self.add_menus_to_environment(app)

    def add_menus_to_environment(self, app: PyonirApp):
        app.TemplateEnvironment.globals['navigation'] = self.menus.get(app.name)


    def build_plugins_navigation(self, app: PyonirApp):
        if app.available_plugins:
            for plgn in app.available_plugins:
                if not hasattr(plgn, 'pages_dirpath'): continue
                self.build_navigation(plgn)
                pass

    def build_navigation(self, app: PyonirApp):
        from pyonir.utilities import get_all_files_from_dir
        from pyonir import Site
        if app is None: return None
        assert hasattr(app, 'pages_dirpath'), "Get menus 'app' parameter does not have a pages dirpath property"
        menus = {}
        submenus = {}
        file_list = get_all_files_from_dir(app.pages_dirpath, app_ctx=app.app_ctx, entry_type=Menu)  # return files using menu schema model

        for menu in file_list:
            # menu = self.schemas.menu.map_input_to_model(pg)
            has_menu = menu.group or menu.parent
            if menu.status == 'hidden' or not menu.url or (not has_menu): continue
            menu.active = self.active_page == menu.url
            if menu.group:
                menus[menu.url] = menu
            elif menu.parent:
                _ref = submenus.get(menu.parent)
                if not _ref:
                    submenus[menu.parent] = [menu]
                else:
                    _ref.append(menu)

        if submenus:
            for k, m in submenus.items():
                pmenu = menus.get(k)
                if not pmenu: continue
                pmenu.sub_menus = m

        res = PyonirCollection(menus.values(), sort_key='rank').group_by('group')
        self.menus[app.name] = res



from dataclasses import dataclass
from pyonir.core import PyonirSchema


@dataclass
class EmailSubscriber(PyonirSchema):
    """
    Represents an email subscriber
    """
    email: str
    subscriptions: list[str]

    def validate_subscriptions(self):
        if not self.subscriptions:
            self._validation_errors.append(f"Subscription is required")

    def validate_email(self):
        import re
        if not re.match(r"[^@]+@[^@]+\.[^@]+", self.email):
            self._validation_errors.append(f"Invalid email address: {self.email}")



from pyonir.pyonir_types import PyonirRequest
from backend.models.EmailSubscriber import EmailSubscriber


async def dynamic_lambda(request: PyonirRequest) -> str:
    return "hello battle rap forever!"

async def demo_items(request: PyonirRequest, sample_id: int=99, version_id: str='0.1'):
    """Home route handler"""
    return f"Main app ITEMS route {sample_id}! {version_id}"

async def subscriber_model(subscriber: EmailSubscriber):
    """Demo takes request body as parameter argument"""
    print(subscriber)
    return subscriber

async def subscriber_values(email: str, subscriptions: list[str]):
    """Demo takes request body as parameter arguments"""
    print(email, subscriptions)
    return f"subscribing {email} to {subscriptions}"

async def some_route(request: PyonirRequest):
    return "hello router annotation"



# -*- coding: utf-8 -*-
import os, sys
from pyonir import utilities
from pyonir.parser import Parsely
from pyonir.core import PyonirHooks, TemplateEnvironment, PyonirRequest, PyonirApp
from pyonir.utilities import dict_to_class, get_attr, process_contents, load_env

# Pyonir settings
PYONIR_DIRPATH = os.path.abspath(os.path.dirname(__file__))
PYONIR_LIBS_DIRPATH = os.path.join(PYONIR_DIRPATH, "libs")
PYONIR_PLUGINS_DIRPATH = os.path.join(PYONIR_LIBS_DIRPATH, 'plugins')
PYONIR_SETUPS_DIRPATH = os.path.join(PYONIR_LIBS_DIRPATH, 'app_setup')
PYONIR_JINJA_DIRPATH = os.path.join(PYONIR_LIBS_DIRPATH, 'jinja')
PYONIR_JINJA_TEMPLATES_DIRPATH = os.path.join(PYONIR_JINJA_DIRPATH, "templates")
PYONIR_JINJA_EXTS_DIRPATH = os.path.join(PYONIR_JINJA_DIRPATH, "extensions")
PYONIR_JINJA_FILTERS_DIRPATH = os.path.join(PYONIR_JINJA_DIRPATH, "filters")
# PYONIR_MESSAGES_FILE = os.path.join(PYONIR_LIBS_DIRPATH, "system-messages.md")
# PYONIR_SSL_KEY = os.path.join(PYONIR_SETUPS_DIRPATH, "content/certs/server.key")
# PYONIR_SSL_CRT = os.path.join(PYONIR_SETUPS_DIRPATH, "content/certs/server.crt")
PYONIR_STATIC_ROUTE = "/pyonir_assets"

__version__: str = '1.0.0'
Site: PyonirApp | None = None

def init(entry_file_path: str, options: dict = None):
    """Initializes existing Pyonir application"""
    global Site
    # Set Global Site instance
    # if options: options = PyonirOptions(**(options or {}))
    sys.path.insert(0, entry_file_path)
    Site = PyonirApp(entry_file_path)
    return Site


from __future__ import annotations
import os
import typing
from typing import Union, Generator, Iterable, Callable, Mapping, get_origin, get_args, get_type_hints, Any
from collections.abc import Iterable as ABCIterable

from pyonir.pyonir_types import PyonirRequest


def is_iterable(tp):
    if not isinstance(tp, Iterable): return False
    origin = get_origin(tp) or tp
    return issubclass(origin, ABCIterable)

def is_generator(tp):
    origin = get_origin(tp) or tp
    return issubclass(origin, Generator)

def is_mappable_type(tp) -> bool:
    origin = get_origin(tp)
    args = get_args(tp)

    # Check if the base is a Mapping (like dict) and it has two type arguments
    return (
        origin is not None and
        issubclass(origin, Mapping) and
        len(args) == 2
    )

def is_scalar_type(tp) -> bool:
    origin = get_origin(tp) or tp
    return origin in (int, float, str, bool)

def is_custom_class(t):
    return t.__init__.__annotations__ #isinstance(t, type) and not t.__module__ == "builtins"

def is_optional_type(t):
    if get_origin(t) is not Union: return t
    return [arg for arg in get_args(t) if arg is not type(None)][0]

def is_callable_type(pt) -> bool:
    return get_origin(pt) is Callable or pt.__name__=='callable'

def cls_mapper(file_obj: any, cls: typing.Callable, from_request: PyonirRequest = None):
    param_name, param_type, param_value = ['','','']
    try:
        if hasattr(cls, '__skip_parsely_deserialization__'): return file_obj
        param_type_map = get_type_hints(cls)
        is_generic = cls.__name__ == 'GenericQueryModel'
        if is_scalar_type(cls):
            return cls(file_obj)
        # if is_generic:
        #     print('isgeneric', file_obj.file_name, file_obj.file_data_type)
        mapper_keys = cls._mapper if hasattr(cls, '_mapper') else {}
        data = get_attr(file_obj, 'data') or {}
        _parsely_data_key = '.'.join(['data', get_attr(cls, '_mapper_key') or cls.__name__.lower()])
        kdata = get_attr(file_obj, _parsely_data_key)
        if kdata: data.update(**kdata)
        cls_args = {}
        res = cls() if is_generic else None
        if hasattr(cls, 'from_dict'): # allows manual mapping of class instance
            return cls.from_dict(file_obj)
        for param_name, param_type in param_type_map.items():
            param_type = is_optional_type(param_type)
            mapper_key = get_attr(mapper_keys, param_name) or param_name
            param_value = get_attr(data, mapper_key) or get_attr(file_obj, mapper_key)
            if param_type == PyonirRequest and from_request:
                param_value = from_request
            if (param_value==param_type) or param_value is None or param_name[0]=='_' or param_name=='return':
                continue
            if is_callable_type(param_type):
                cls_args[mapper_key] = param_value
            elif is_iterable(param_type):
                iter_ptype = get_args(param_type)
                is_mapp = is_mappable_type(param_type)
                if is_mapp:
                    ktype, vtype = iter_ptype
                    is_list = is_iterable(vtype)
                    if is_list: vtype = get_args(vtype)[0]
                    cls_args[param_name] = {ktype(key): [cls_mapper(lval, vtype) for lval in value] if is_list else cls_mapper(value, vtype) for key, value in param_value.items()}
                else:
                    cls_args[param_name] = [cls_mapper(itm, iter_ptype[0]) for itm in param_value]
            else:
                is_instance = param_value == param_type
                is_typed = isinstance(param_value, param_type)
                if is_instance and from_request:
                    param_value = cls_mapper(from_request.form, param_type)
                should_spread = isinstance(param_value, dict)
                use_value = param_type == PyonirRequest and from_request or is_typed or is_instance
                v = param_value if use_value else param_type(**param_value) if should_spread else param_type(param_value)
                cls_args[param_name] = v

        if from_request: return cls_args
        if not from_request and param_type_map: res = cls(**cls_args)
        # setattr(res, '@model', cls.__name__)

        if is_generic:
            # lookup values for class instance
            for key in cls.__dict__.keys():
                if key[0]=='_': continue
                value = get_attr(data, key) or get_attr(file_obj, key)
                setattr(res, key, value)
        if hasattr(cls, '_mapper_merge'):
            # Sets non class attributes onto the model instance
            for key, value in data.items():
                if isinstance(getattr(cls, key, None), property): continue # skip active attributes
                if param_type_map.get(key) or key[0]=='_': continue # skip private attributes
                setattr(res, key, value)

    except Exception as e:
        print(f"Cls Mapper failed to create a {cls.__name__} instance due to map '{param_name}' parameter wasn't a type of {param_type} : {type(param_value)}")
        raise

    return res


def process_contents(path, app_ctx=None, file_model: any = None):
    """Deserializes all files within the contents directory"""
    from pyonir.parser import Parsely

    def update(self):
        for key in self.__dict__.keys():
            val = get_attr(self, key)
            if not hasattr(val, '_path'): continue
            pval = Parsely(val._path, app_ctx)
            setattr(self, key, dict_to_class({**pval.data, '_path': pval.file_path}, key))
        pass

    key = os.path.basename(path)
    res = type(key, (object,), {})() # generic map
    pgs = get_all_files_from_dir(path, app_ctx=app_ctx, entry_type=file_model)
    for pg in pgs:
        name = getattr(pg, 'file_name')
        setattr(res, name, pg)
    return res


def dict_to_class(data: dict, name: str = 'T'):
    """
    Converts a dictionary into a class object with the given name.

    Args:
        data (dict): The dictionary to convert.
        name (str): The name of the class.

    Returns:
        object: An instance of the dynamically created class with attributes from the dictionary.
    """
    # Dynamically create a new class
    cls = type(name, (object,), {})

    # Create an instance of the class
    instance = cls()

    # Assign dictionary keys as attributes of the instance
    for key, value in data.items():
        if isinstance(getattr(cls, key, None), property): continue
        # if isinstance(value, dict):
        #     value = dict_to_class(value, key)
        setattr(instance, key, value)

    return instance

def get_attr(rowObj, attrPath=None, default=None, rtn_none=True):
    """
    Resolves nested attribute or dictionary key paths.

    Args:
        obj: the root object
        attr_path: dot-separated string or list for nested access
        default: fallback value if the target is None or missing
        return_none: if True, returns `None` on missing keys/attrs instead of the original object

    Returns:
        The nested value, or `default`, or `obj` based on fallback rules.
    """
    if attrPath == None: return rowObj
    attrPath = attrPath if isinstance(attrPath, list) else attrPath.split('.')
    targetObj = None
    for key in attrPath:
        try:
            if targetObj:
                targetObj = targetObj[key]
            else:
                targetObj = rowObj.get(key)
            pass
        except (KeyError, AttributeError, TypeError) as e:
            if targetObj:
                targetObj = getattr(targetObj, key, None)
            else:
                targetObj = getattr(rowObj, key, None)
            pass
    if targetObj is None and rtn_none:
        return default or None

    return targetObj


def remove_html_tags(text):
    """Remove html tags from a excerpt string"""
    import re
    clean_style = re.sub(r'<style.*?</style>', '', text, flags=re.DOTALL)
    clean_html = re.sub(re.compile('<.*?>'), '', clean_style)
    return clean_html.replace('\n', ' ')


def camel_to_snake(camel_str):
    """Converts camelcase into snake case. Thanks Chat GPT"""
    import re
    snake_str = re.sub(r'(?<!^)(?=[A-Z])', '_', camel_str).lower()
    return snake_str


def deserialize_datestr(datestr, timestr="00:00", fmt="%Y-%m-%d %H:%M:%S %p", zone="US/Eastern", auto_correct=True):
    from datetime import datetime
    import pytz
    if not isinstance(datestr, str): return datestr

    def correct_format(date_str):
        try:
            date_str = date_str.strip().replace('/', '-')
            date_str, _, timestr = date_str.partition(" ")
            timestr = timestr if timestr != "" else '12:13:14 AM'
            has_period = timestr.endswith("M")
            if not has_period:
                timestr += " AM"
            y, m, *d = date_str.split("-")
            d = "".join(d)
            fdate = f"{y}-{int(m):02d}-{d}"
            if int(y) < int(d):
                fdate = f"{d}-{int(y):02d}-{m}"
                print(f"\tIncorrect format on date string {date_str}. it should be {fdate}")
                return fdate
            return f"{fdate} {timestr}"
        except Exception as e:
            return None

    try:
        return pytz.utc.localize(datetime.strptime(datestr, fmt))
    except ValueError as e:
        # return str(e)
        if not auto_correct: return str(e)
        datestr = correct_format(datestr)
        return pytz.utc.localize(datetime.strptime(datestr, fmt)) if datestr else None


def sortBykey(listobj, sort_by_key="", limit="", reverse=True):
    """Sorts list of obj by key"""

    def get_path_object(rowObj, path):
        targetObj = None
        for key in path.split('.'):
            try:
                if targetObj:
                    targetObj = targetObj[key]
                else:
                    targetObj = rowObj[key]
                pass
            except Exception as error:
                raise error
        return targetObj

    try:
        sorted_dict = sorted(getattr(listobj, 'data', listobj), key=lambda obj: get_path_object(obj, sort_by_key),
                             reverse=reverse)
        # sorted_dict = sorted(getattr(listobj,'data', listobj), key = lambda x:x[sort_by_key], reverse=reverse)
        if limit:
            return sorted_dict[:limit]
        return sorted_dict
    except Exception as e:
        return listobj


def get_all_files_from_dir(abs_dirpath: str,
                           app_ctx: list = None,
                           entry_type: any = None,
                           include_only: str = None,
                           exclude_dirs: list[str] = None,
                           exclude_file: str = None,
                           force_all: bool = True) -> Generator:
    """Returns a generator of files from a directory path"""

    from pyonir.parser import ALLOWED_CONTENT_EXTENSIONS, IGNORE_FILES,Page, Parsely, ParselyMedia
    from pyonir.core import PyonirBase
    if abs_dirpath in (exclude_dirs or []): return []
    if exclude_file is None: exclude_file = []
    _, _, pages_dirpath, _ = app_ctx

    def get_datatype(parentdir, rel_filepath):
        filepath = os.path.normpath(os.path.join(parentdir, rel_filepath))
        if entry_type == 'path':
            return filepath
        ismedia = not filepath.endswith(ALLOWED_CONTENT_EXTENSIONS)
        generic_model = entry_type if entry_type and entry_type.__name__=='GenericQueryModel' else None
        pf = Parsely(filepath, app_ctx, generic_model)
        if ismedia:
            return cls_mapper(pf, ParselyMedia)
        return cls_mapper(pf, entry_type or Page ) if entry_type or pf.is_page else pf.map_to_model(None)


    def should_skip(parentdir, filename):
        parentdir = parentdir.replace(pages_dirpath, "").lstrip(os.path.sep)
        is_hidden_dir = parentdir.startswith(IGNORE_FILES)
        is_media_file = filename.endswith(PyonirBase.MEDIA_EXTENSIONS)
        is_invalid_file = not is_media_file and not filename.endswith(ALLOWED_CONTENT_EXTENSIONS)
        is_ignored_file = filename in (IGNORE_FILES + tuple(exclude_file))
        is_private_file = filename.startswith(IGNORE_FILES)
        if filename!='.routes.md' and force_all and not is_invalid_file: return False
        if is_hidden_dir or is_invalid_file or is_ignored_file or is_private_file: return True

    for parentdir, subs, files in os.walk(os.path.normpath(abs_dirpath)):
        folderRoot = parentdir.replace(pages_dirpath, "").lstrip(os.path.sep)
        subFolderRoot = os.path.basename(folderRoot)
        skipRoot = (folderRoot in IGNORE_FILES
                    or folderRoot.startswith(IGNORE_FILES)
                    or subFolderRoot.startswith(IGNORE_FILES))
        skipSubs = subFolderRoot in exclude_dirs if exclude_dirs else 0
        if skipRoot or skipSubs: continue

        for filename in files:
            # include_only_file = (include_only and filename != include_only)
            if should_skip(parentdir, filename): continue
            # if filename in exclude_file or include_only_file: continue
            # if not force_all and not filename.endswith(ALLOWED_CONTENT_EXTENSIONS): continue
            # if not should_skip(parentdir, filename) or filename in IGNORE_FILES: continue
            yield get_datatype(parentdir, filename)


def delete_file(full_filepath):
    import shutil
    if os.path.isdir(full_filepath):
        shutil.rmtree(full_filepath)
        return True
    elif os.path.isfile(full_filepath):
        os.remove(full_filepath)
        return True
    return False


def create_file(file_abspath: str, data: any = None, is_json: bool = False, mode='w') -> bool:
    def write_file(file_abspath, data, is_json=False, mode='w'):
        import json
        with open(file_abspath, mode, encoding="utf-8") as f:
            if is_json:
                json.dump(data, f, indent=2, sort_keys=True, default=json_serial)
            else:
                f.write(data)

    """Creates a new file based on provided data
    Args:
        file_abspath: str = path to proposed file
        data: any = contents to write into file
        is_json: bool = strict json file
        mode: str = write mode for file w|w+|a
    Returns:
        bool: The return value if file was created successfully
    """
    if not os.path.exists(os.path.dirname(file_abspath)):
        os.makedirs(os.path.dirname(file_abspath))
    try:

        if is_json:
            file_abspath = file_abspath.replace(".md", ".json")
        write_file(file_abspath, data, is_json=is_json, mode=mode)

        return True
    except Exception as e:
        print(f"Error create_file method: {str(e)}")
        return False


def copy_assets(src: str, dst: str, purge: bool = True):
    """Copies files from a source directory into a destination directory with option to purge destination"""
    import shutil
    from shutil import ignore_patterns
    print(f"{PrntColrs.OKBLUE}Coping `{src}` resource into {dst}{PrntColrs.RESET}")
    try:
        if os.path.exists(dst) and purge:
            shutil.rmtree(dst)
        if os.path.isfile(src):
            shutil.copyfile(src, dst)
        if os.path.isdir(src):
            shutil.copytree(src, dst, ignore=ignore_patterns('__pycache__', '*.pyc', 'tmp*', 'node_modules', '.*'))
    except Exception as e:
        raise



def json_serial(obj):
    """JSON serializer for nested objects not serializable by default jsonify"""
    from datetime import datetime
    from .parser import Parsely
    if isinstance(obj, datetime):
        return obj.isoformat()
    elif isinstance(obj, Generator) or hasattr(obj, 'mapping'):
        return list(obj)
    elif isinstance(obj, Parsely):
        return obj.data
    elif hasattr(obj, 'to_json'):
        return obj.to_json()
    else:
        return obj if not hasattr(obj, '__dict__') else obj.__dict__


def load_modules_from(pkg_dirpath, as_list: bool = False, only_packages:bool = False)-> dict[Any, Any] | list[Any]:
    loaded_mods = {} if not as_list else []
    loaded_funcs = {} if not as_list else []
    if not os.path.exists(pkg_dirpath): return loaded_funcs
    for mod_file in os.listdir(pkg_dirpath):
        name,_, ext = mod_file.partition('.')
        if only_packages:
            pkg_abspath = os.path.join(pkg_dirpath, mod_file, '__init__.py')
            mod, func = get_module(pkg_abspath, name)
        else:
            if ext!='py': continue
            mod_abspath = os.path.join(pkg_dirpath, name.strip())+'.py'
            mod, func = get_module(mod_abspath, name)
        if as_list:
            loaded_funcs.append(func)
        else:
            loaded_mods[name] = mod
            loaded_funcs[name] = func

    return loaded_funcs

def get_module(pkg_path: str, callable_name: str) -> tuple[typing.Any, typing.Callable]:
    import sys, importlib
    spec = importlib.util.spec_from_file_location(callable_name, pkg_path)
    if spec is None or spec.loader is None:
        raise ImportError(f"Could not load spec for {callable_name} from {pkg_path}")

    module = importlib.util.module_from_spec(spec)
    # sys.modules[callable_name] = module  # Optional: allows intra-module imports to resolve
    spec.loader.exec_module(module)

    func = get_attr(module, callable_name) or get_attr(module, module.__name__)
    return module, func

# def _get_module(pkg_path: str, callable_name: str) -> tuple[any, typing.Callable]:
#     from importlib import util
#     mod = util.spec_from_file_location(callable_name, pkg_path).loader.load_module(callable_name)
#     func = getattr(mod, callable_name)
#
#     return (mod, func)

def generate_id():
    import uuid
    return str(uuid.uuid1())

def generate_base64_id(value):
    import base64
    return base64.b64encode(value.encode('utf-8'))

def load_env(path=".env") -> object:
    from collections import defaultdict
    if not os.path.exists(path): return {}
    def set_nested(d, keys, value):
        """Helper to set value in nested dictionary using dot-separated keys."""
        for key in keys[:-1]:
            d = d.setdefault(key, {})
        d[keys[-1]] = value

    env_data = defaultdict(dict)

    with open(path) as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            if "=" not in line:
                continue

            key, value = line.split("=", 1)
            key = key.strip()
            value = value.strip()

            # Set in os.environ (flat)
            os.environ.setdefault(key, value)

            # Set in nested dict (structured)
            keys = key.split(".")
            set_nested(env_data, keys, value)

    return dict_to_class(env_data, 'env')


class PrntColrs:
    RESET = '\033[0m'
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\x1b[0;92;49m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'



import os, json, textwrap
from pyonir.parser import Parsely
from pyonir import init

def generate_pyonir_types():
    from pyonir.core import PyonirApp, PyonirRequest, PyonirPlugin

    for cls in [PyonirApp, PyonirRequest, PyonirPlugin]:
        generate_dataclass_from_class(cls)

def generate_dataclass_from_class(cls, output_dir="types"):
    from typing import get_type_hints
    attr_map = get_type_hints(cls)
    props_map = {k: type(v).__name__ for k, v in cls.__dict__.items() if isinstance(v, property)}
    meth_map = {k: callable for k, v in cls.__dict__.items() if callable(v)}
    all_map = dict(**props_map, **meth_map, **attr_map)
    lines = [f"class {cls.__name__}:"]
    if not cls.__annotations__:
        lines.append("    pass")
    else:
        for name, typ in all_map.items():
            lines.append(f"    {name}: {typ.__class__.__name__}")
    with open(os.path.join(os.path.dirname(__file__), output_dir, f"{cls.__name__}.py"), "w") as f:
        f.write("\n".join(lines))

def generate_tests(parsely: Parsely):
    cases = []
    name = parsely.__class__.__name__
    space = "\t"
    for key, value in parsely.data.items():
        test_case = textwrap.dedent(f"""
        {space}def test_{key}(self):
            {space}self.assertEqual({json.dumps(value)}, self.parselyFile.data.get('{key}'))
        """)
        cases.append(test_case)

    case_meths = "\n".join(cases)
    test_class = textwrap.dedent(f"""\
import unittest, os
true = True
class {name}Tests(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        from pyonir.parser import Parsely
        from pyonir import init
        App = init(__file__)
        cls.parselyFile = Parsely(os.path.join(os.path.dirname(__file__), 'test.md'), App.app_ctx)
    {case_meths}
    """)

    parsely.save(os.path.join(os.path.dirname(__file__), 'generated_test.py'), test_class)

if __name__=='__main__':
    # generate_pyonir_types()
    App = init(__file__)
    # file = App.parsely_file('test.md')
    # generate_tests(file)
    # print(file.data)
    pass


from __future__ import annotations
import typing, os

from jinja2 import Environment
from starlette.requests import Request as StarletteRequest

from pyonir.pyonir_types import PyonirServer, Theme, PyonirHooks, Parsely, ParselyPagination, AppRequestPaths, AppCtx, \
    PyonirRouters, RoutePath

# Environments
DEV_ENV:str = 'LOCAL'
STAGE_ENV:str = 'STAGING'
PROD_ENV:str = 'PROD'

TEXT_RES: str = 'text/html'
JSON_RES: str = 'application/json'
EVENT_RES: str = 'text/event-stream'
PAGINATE_LIMIT: int = 6


class PyonirSchema:
    # default_file_attributes = ['file_name','file_path','file_dirname','file_data_type','file_ctx','file_created_on']

    """Schema class enables validation when model class initializes"""
    # Schemas configs
    # PROTECTED_FIELD_PREFIX = '@'
    # # Private fields should can be read and written but should not be exposed
    # PRIVATE_FIELDS = ('password', 'ssn', 'auth_token', 'id', 'uid', 'privateKey')
    # # Protected fields are allowed to be read but are not allowed to be changed directly
    # PROTECTED_FIELDS = (
    #                        'created_on', 'modified_on', 'modified_by', 'last_modified_by',
    #                        'date_created', 'date_modified', 'raw', 'provider_model') + PRIVATE_FIELDS

    def validate(self):
        """validates a given property with accessible validation method"""
        for name, value in self.__dict__.items():
            if name.startswith('_'): continue
            validator_fn = getattr(self, f'validate_{name}', None)
            if validator_fn: validator_fn()
        pass

    def __post_init__(self):
        self._validation_errors = []
        self.validate()

    @staticmethod
    def generic_query_model(model_fields: str):
        if not model_fields: return None
        from pyonir.parser import Parsely
        mapper = {}
        params = {"_mapper": mapper}
        for k in Parsely.default_file_attributes+model_fields.split(','):
            if ':' in k:
                k,_, src = k.partition(':')
                mapper[k] = src
            params[k] = None
        return type('GenericQueryModel', (object,), params)

class PyonirCollection:
    SortedList = None
    get_attr =  None
    dict_to_class = None

    def __init__(self, items: typing.Iterable, sort_key: str = None):
        from sortedcontainers import SortedList
        from pyonir.utilities import get_attr, dict_to_class
        self.SortedList = SortedList
        self.get_attr = get_attr
        self.dict_to_class = dict_to_class

        self._query_path = ''
        key = lambda x: self.get_attr(x, sort_key or 'file_created_on')
        try:
            items = list(items)
            self.collection = SortedList(items, key=key)
        except Exception as e:
            raise

    @staticmethod
    def coerce_bool(value: str):
        d = ['false', 'true']
        try:
            i = d.index(value.lower().strip())
            return True if i else False
        except ValueError as e:
            return value.strip()

    @staticmethod
    def parse_params(param: str):
        k, _, v = param.partition(':')
        op = '='
        is_eq = lambda x: x[1]==':'
        if v.startswith('>'):
            eqs = is_eq(v)
            op = '>=' if eqs else '>'
            v = v[1:] if not eqs else v[2:]
        elif v.startswith('<'):
            eqs = is_eq(v)
            op = '<=' if eqs else '<'
            v = v[1:] if not eqs else v[2:]
            pass
        else:
            pass
        # v = True if v.strip()=='true' else v.strip()
        return {"attr": k.strip(), "op":op, "value":PyonirCollection.coerce_bool(v)}

    @classmethod
    def query(cls, query_path: str,
             app_ctx: PyonirApp = None,
             data_model: any = None,
             include_only: str = None,
             exclude_dirs: list[str] = None,
             exclude_file: list[str] = None,
             force_all: bool = True,
              sort_key: str = None):
        """queries the file system for list of files"""
        from pyonir.utilities import get_all_files_from_dir
        gen_data = get_all_files_from_dir(query_path, app_ctx=app_ctx, entry_type=data_model, include_only=include_only,
                                          exclude_dirs=exclude_dirs, exclude_file=exclude_file, force_all=force_all)
        return cls(gen_data, sort_key=sort_key)

    def prev_next(self, input_file: Parsely):
        """Returns the previous and next files relative to the input file"""

        prv = None
        nxt = None
        pc = self.query(input_file.file_dirpath)
        pc.collection = iter(pc.collection)
        for cfile in pc.collection:
            if cfile.file_status == 'hidden': continue
            if cfile.file_path == input_file.file_path:
                nxt = next(pc.collection, None)
                break
            else:
                prv = cfile
        return self.dict_to_class({"next": nxt, "prev": prv})

    def find(self, value: any, from_attr: str = 'file_name'):
        """Returns the first item where attr == value"""
        return next((item for item in self.collection if getattr(item, from_attr, None) == value), None)

    def where(self, attr, op="=", value=None):
        """Returns a list of items where attr == value"""
        # if value is None:
        #     # assume 'op' is actually the value if only two args were passed
        #     value = op
        #     op = "="

        def match(item):
            actual = self.get_attr(item, attr)
            if not hasattr(item, attr):
                return False
            if actual and not value:
                return True # checking only if item has an attribute
            elif op == "=":
                return actual == value
            elif op == "in" or op == "contains":
                return actual in value if actual is not None else False
            elif op == ">":
                return actual > value
            elif op == "<":
                return actual < value
            elif op == ">=":
                return actual >= value
            elif op == "<=":
                return actual <= value
            elif op == "!=":
                return actual != value
            return False
        if isinstance(attr, typing.Callable): match = attr
        return PyonirCollection(filter(match, list(self.collection)))

    def paginate(self, start: int, end: int, reversed: bool = False):
        """Returns a slice of the items list"""
        sl = self.collection.islice(start, end, reverse=reversed) if end else self.collection
        return sl #self.collection[start:end]

    def group_by(self, key: str | typing.Callable):
        """
        Groups items by a given attribute or function.
        If `key` is a string, it will group by that attribute.
        If `key` is a function, it will call the function for each item.
        """
        from collections import defaultdict
        grouped = defaultdict(list)

        for item in self.collection:
            k = key(item) if callable(key) else getattr(item, key, None)
            grouped[k].append(item)

        return dict(grouped)

    def paginated_collection(self, query_params=None)-> ParselyPagination | None:
        """Paginates a list into smaller segments based on curr_pg and display limit"""
        if query_params is None: query_params = {}
        from pyonir import Site
        if not Site: return None
        from pyonir.core import ParselyPagination
        request: PyonirRequest = Site.TemplateEnvironment.globals['request']
        if not hasattr(request, 'limit'): return None
        req_pg = self.get_attr(request.query_params, 'pg') or 1
        limit = query_params.get('limit', request.limit)
        curr_pg = int(query_params.get('pg', req_pg)) or 1
        sort_key = query_params.get('sort_key')
        where_key = query_params.get('where')
        if sort_key:
            self.collection = self.SortedList(self.collection, lambda x: self.get_attr(x, sort_key))
        if where_key:
            where_key = [PyonirCollection.parse_params(ex) for ex in where_key.split(',')]
            self.collection = self.where(**where_key[0])
        force_all = limit=='*'

        max_count = len(self.collection)
        limit = 0 if force_all else int(limit)
        page_num = 0 if force_all else int(curr_pg)
        start = (page_num * limit) - limit
        end = (limit * page_num)
        pg = (max_count // limit) + (max_count % limit > 0) if limit > 0 else 0

        pag_data = self.paginate(start=start, end=end, reversed=True) if not force_all else self.collection

        return ParselyPagination(**{
            'curr_page': page_num,
            'page_nums': [n for n in range(1, pg + 1)] if pg else None,
            'limit': limit,
            'max_count': max_count,
            'items': list(pag_data)
        })

    def __len__(self):
        return self.collection._len

    def __iter__(self):
        return iter(self.collection)


class PyonirRequest:

    def __init__(self, server_request: StarletteRequest):
        from pyonir.utilities import get_attr

        self.server_response = None
        self.file: typing.Optional[Parsely] = None
        self.server_request: StarletteRequest = server_request
        self.raw_path = "/".join(str(self.server_request.url).split(str(self.server_request.base_url)))
        self.method = self.server_request.method
        self.path = self.server_request.url.path
        self.path_params = dict(self.server_request.path_params)
        self.url = f"{self.path}"
        self.slug = self.path.lstrip('/').rstrip('/')
        self.query_params = self.get_params(self.server_request.url.query)
        self.parts = self.slug.split('/') if self.slug else []
        self.limit = get_attr(self.query_params, 'limit', PAGINATE_LIMIT)
        self.model = get_attr(self.query_params, 'model')
        self.is_home = (self.path == '')
        self.is_api = False
        self.form = {}
        self.files = []
        self.ip = self.server_request.client.host
        self.host = str(self.server_request.base_url).rstrip('/')
        self.protocol = self.server_request.scope.get('type') + "://"
        self.headers = PyonirRequest.process_header(self.server_request.headers)
        self.browser = self.headers.get('user-agent', '').split('/').pop(0) if self.headers else "UnknownAgent"
        if self.slug.startswith('api'): self.headers['accept'] = JSON_RES
        self.type: TEXT_RES | JSON_RES | EVENT_RES = self.headers.get('accept')
        self.status_code: int = 200
        self.app_ctx_name: str = ''

    @property
    def redirect(self):
        return self.form.get('redirect', self.form.get('redirect_to'))

    async def process_request_data(self):
        """Get form data and file upload contents from request"""

        from pyonir import Site
        import json

        def secure_upload_filename(filename):
            import re
            # Strip leading and trailing whitespace from the filename
            filename = filename.strip()

            # Replace spaces with underscores
            filename = filename.replace(' ', '_')

            # Remove any remaining unsafe characters using a regular expression
            # Allow only alphanumeric characters, underscores, hyphens, dots, and slashes
            filename = re.sub(r'[^a-zA-Z0-9_.-]', '', filename)

            # Ensure the filename doesn't contain multiple consecutive dots (.) or start with one
            filename = re.sub(r'\.+', '.', filename).lstrip('.')

            # Return the filename as lowercase for consistency
            return filename.lower()

        try:
            try:
                ajson = await self.server_request.json()
                if isinstance(ajson, str): ajson = json.loads(ajson)
                self.form.update(ajson)
            except Exception as ee:
                # multipart/form-data
                form = await self.server_request.form()
                files = []
                for name, content in form.multi_items():
                    if name == 'files':
                        # filedata = await content.read()
                        mediaFile = (secure_upload_filename(content.filename), content, Site.uploads_dirpath)
                        self.files.append(mediaFile)
                    else:
                        if self.form.get(name): # convert form name into a list
                            currvalue = self.form[name]
                            if isinstance(currvalue, list):
                                currvalue.append(content)
                            else:
                                self.form[name] = [currvalue, content]
                        else:
                            self.form[name] = content
        except Exception as e:
            raise

    def derive_status_code(self, is_router_method: bool):
        """Create status code for web request based on a file's availability, status_code property"""
        from pyonir.parser import ParselyFileStatus

        code = 404
        if self.file.file_status == ParselyFileStatus.PROTECTED:
            self.file.data = {'template': '40x.html', 'content': f'Unauthorized access to this resource.', 'url': self.url, 'slug': self.slug}
            code = 401
        elif self.file.file_status == ParselyFileStatus.PUBLIC or is_router_method:
            code = 200
        self.status_code = code #200 if self.file.file_exists or is_router_method else 404

    def render_error(self):
        """Data output for an unknown file path for a web request"""
        return {
            "url": self.url,
            "method": self.method,
            "status": self.status_code,
            "res": self.server_response,
            "title": f"{self.path} was not found!",
            "content": f"Perhaps this page once lived but has now been archived or permanently removed from {self.app_ctx_name}."
        }

    def resolve_request_to_file(
        self,
        path_str: str,
        app: PyonirApp,
        skip_vanity: bool = False
    ) -> typing.Tuple[PyonirApp, typing.Optional[Parsely]]:
        """Resolve a request URL to a file on disk, checking plugin paths first, then the main app."""
        from pyonir import Site
        from pyonir.parser import Parsely
        is_home = path_str == '/'

        # First, check plugins if available and not home
        if not is_home and hasattr(app, 'plugins_activated'):
            for plugin in app.plugins_activated:
                if not plugin.request_paths: continue
                resolved_app, parsed = self.resolve_request_to_file(path_str, plugin)
                if parsed and parsed.file_exists:
                    return resolved_app, parsed

        ctx_route, ctx_paths = app.request_paths or ('', [])
        ctx_route = ctx_route or ''
        ctx_slug = ctx_route[1:]
        path_slug = path_str[1:]

        # Normalize API prefix and path segments
        if path_str.startswith('/api'):
            path_str = path_str.replace('/api', '')

        request_segments = [
            segment for segment in path_slug.split('/')
            if segment and segment not in ('api', ctx_slug)
        ]

        # Skip if no paths or route doesn't match
        if not ctx_paths or (not is_home and not path_str.startswith(ctx_route)):
            return app, None

        # Try resolving to actual file paths
        protected_segment = [s if i > len(request_segments)-1 else f'_{s}' for i,s in enumerate(request_segments)]
        for root_path in ctx_paths:
            category_index = os.path.join(root_path, *request_segments, 'index.md')
            single_page = os.path.join(root_path, *request_segments) + Site.EXTENSIONS['file']
            single_protected_page = os.path.join(root_path, *protected_segment) + Site.EXTENSIONS['file']

            for candidate in (category_index, single_page, single_protected_page):
                if os.path.exists(candidate):
                    return app, Parsely(candidate, app.app_ctx)

        # Fallback: check for .routes.md inside pages_dirpath
        if hasattr(app, 'pages_dirpath'):
            fallback_route = os.path.join(app.pages_dirpath, '.routes.md')
            if os.path.exists(fallback_route):
                return app, Parsely(fallback_route, app.app_ctx)

        return app, Parsely('404_ERROR', app.app_ctx)

    @staticmethod
    def process_header(headers):
        nheaders = dict(headers)
        nheaders['accept'] = nheaders.get('accept', TEXT_RES).split(',', 1)[0]
        agent = nheaders.get('user-agent', '')
        nheaders['user-agent'] = agent.split(' ').pop().split('/', 1)[0]
        return nheaders

    @staticmethod
    def get_params(url):
        import urllib
        from pyonir.utilities import dict_to_class
        args = {params.split('=')[0]: urllib.parse.unquote(params.split("=").pop()) for params in
                url.split('&') if params != ''}
        if args.get('model'): del args['model']
        return dict_to_class(args, 'query_params')


class PyonirBase:
    """Pyonir Base Application Configs"""
    pyonir_path: str = os.path.dirname(__file__)
    endpoint: str | None = None
    # Default config settings
    EXTENSIONS = {"file": ".md", "settings": ".json"}
    THUMBNAIL_DEFAULT = (230, 350)
    PROTECTED_FILES = {'.', '_', '<', '>', '(', ')', '$', '!', '._'}
    IGNORE_FILES = {'.vscode', '.vs', '.DS_Store', '__pycache__', '.git'}

    PAGINATE_LIMIT: int = 6
    DATE_FORMAT: str = "%Y-%m-%d %I:%M:%S %p"
    TIMEZONE: str = "US/Eastern"
    # ALLOWED_UPLOAD_EXTENSIONS: set[str] = {'jpg', 'JPG', 'PNG', 'png', 'txt', 'md', 'jpeg', 'pdf', 'svg', 'gif'}
    MEDIA_EXTENSIONS = (
        # Audio
        ".mp3", ".wav", ".aac", ".flac", ".ogg", ".m4a", ".wma", ".aiff", ".alac",

        # Video
        ".mp4", ".mov", ".avi", ".mkv", ".webm", ".flv", ".wmv", ".mpeg", ".mpg", ".3gp",

        # Images
        ".jpg", ".jpeg", ".png", ".gif", ".bmp", ".tiff", ".webp", ".svg", ".heic",

        # Raw Image Formats
        ".raw", ".cr2", ".nef", ".orf", ".arw", ".dng",

        # Media Playlists / Containers
        ".m3u", ".m3u8", ".pls", ".asx", ".m4v", ".ts"
    )
    # Base application  default directories
    # Overriding these properties will dynamicall change path properties
    SOFTWARE_VERSION: str = '' # pyonir version number
    APPS_DIRNAME: str = "apps"  # dirname for any child apps
    BACKEND_DIRNAME: str = "backend"  # dirname for all backend python files
    FRONTEND_DIRNAME: str = "frontend"  # dirname for all themes, jinja templates, html, css, and js
    CONTENTS_DIRNAME: str = "contents"  # dirname for site parsely file data
    THEMES_DIRNAME: str = "themes"  # dirname for site themes
    CONFIGS_DIRNAME: str = 'configs'
    TEMPLATES_DIRNAME: str = 'templates'
    SSG_DIRNAME: str = 'static_site'
    SSG_IN_PROGRESS: bool = False

    # Contents sub directory default names
    UPLOADS_THUMBNAIL_DIRNAME: str = "thumbnails" # resized image directory name
    UPLOADS_DIRNAME: str = "uploads" # url name for serving uploaded assets
    ASSETS_DIRNAME: str = "public" # url name for serving static assets css and js
    API_DIRNAME: str = "api" # directory for serving API endpoints and resolver routes
    PAGES_DIRNAME: str = "pages" # directory for serving HTML endpoints with file based routing
    CONFIG_FILENAME: str = "app" # main application configuration file name within contents/configs directory

    # Application paths
    app_dirpath: str = '' # directory path to site's main.py file
    app_name: str = '' # directory name for application main.py file
    app_account_name: str = '' # parent directory from the site's root directory (used for multi-site configurations)

    # Application routes
    API_ROUTE = f"/{API_DIRNAME}"  # Api base path for accessing pages as JSON
    ASSETS_ROUTE = f"/{ASSETS_DIRNAME}"  # serves static assets from configured theme
    UPLOADS_ROUTE = f"/{UPLOADS_DIRNAME}"  # Upload base path to access resources within upload directory
    TemplateEnvironment: TemplateEnvironment = None # Template environment configurations
    routing_paths: list[RoutePath] | None = []

    @property
    def request_paths(self) -> AppRequestPaths:
        return self.endpoint, self.routing_paths

    @property
    def module(self):
        """The application module directory name"""
        return self.__module__.split('.').pop()

    @staticmethod
    def generate_resolvers(cls: callable, output_dirpath: str, namespace: str):
        """Automatically generate api endpoints from service class."""
        import textwrap
        from pyonir.utilities import create_file

        resolver_template = textwrap.dedent("""\
        @resolvers:
            GET.call: {call_path}
        ===
        {docs}
        """).strip()
        name = cls.__class__.__name__
        endpoint_meths = [a for a in dir(cls) if not a.startswith('_') and hasattr(getattr(cls, a), '__call__')]
        formatted_cls_name = name[0].lower()+name[1:]
        print(f"Generating {name} API endpoint definitions for:")
        for meth_name in endpoint_meths:
            file_path = os.path.join(output_dirpath, meth_name+'.md')
            if os.path.exists(file_path): continue
            meth: callable = getattr(cls, meth_name)
            namespace_instance_path = f"{namespace}.{formatted_cls_name}.{meth_name}"
            docs = textwrap.dedent(meth.__doc__).strip() if meth.__doc__ else ''
            m_temp = resolver_template.format(call_path=namespace_instance_path, docs=docs)
            create_file(file_path, m_temp)
            print(f"\t{meth_name} at {file_path}")

    def generate_static_website(self):
        """Generates Static website into the specified static_site_dirpath"""
        import time
        from pyonir.server import generate_nginx_conf
        from pyonir import utilities

        self.SSG_IN_PROGRESS = True
        count = 0
        print(f"{utilities.PrntColrs.OKBLUE}1. Coping Assets")
        try:
            self.run([])
            site_map_path = os.path.join(self.ssg_dirpath, 'sitemap.xml')
            # generate_nginx_conf(self)
            print(f"{utilities.PrntColrs.OKCYAN}3. Generating Static Pages")

            self.TemplateEnvironment.globals['is_ssg'] = True
            start_time = time.perf_counter()

            all_pages = utilities.get_all_files_from_dir(self.pages_dirpath, app_ctx=self.app_ctx)
            xmls = []
            for page in all_pages:
                self.TemplateEnvironment.globals['request'] = page  # pg_req
                count += page.generate_static_file()
                t = f"<url><loc>{self.protocol}://{self.domain}{page.url}</loc><priority>1.0</priority></url>\n"
                xmls.append(t)
                self.TemplateEnvironment.block_pull_cache.clear()

            # Compile sitemap
            smap = f'<?xml version="1.0" encoding="UTF-8"?><urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9"><url><loc>{self.domain}</loc><priority>1.0</priority></url> {"".join(xmls)} </urlset>'
            utilities.create_file(site_map_path, smap, 0)

            # Copy pyonir static assets for js and css vendor libraries into ssg directory
            # utilities.copy_assets(PYONIR_STATIC_DIRPATH, os.path.join(self.ssg_dirpath, PYONIR_STATIC_ROUTE.lstrip('/')))

            # Copy theme static css, js files into ssg directory
            utilities.copy_assets(self.TemplateEnvironment.themes.active_theme.static_dirpath, os.path.join(self.ssg_dirpath, self.ASSETS_DIRNAME))

            end_time = time.perf_counter() - start_time
            ms = end_time * 1000
            count += 3
            msg = f"SSG generated {count} html/json files in {round(end_time, 2)} secs :  {round(ms, 2)} ms"
            print(f'\033[95m {msg}')
        except Exception as e:
            msg = f"SSG encountered an error: {str(e)}"
            raise

        self.SSG_IN_PROGRESS = False
        response = {"status": "COMPLETE", "msg": msg, "files": count}
        print(response)
        print(utilities.PrntColrs.RESET)
        return response


class PyonirPlugin(PyonirBase):

    def __init__(self, app: PyonirApp, app_entrypoint: str = None):
        self._app_ctx = None
        self.app: PyonirApp = app
        self.app_entrypoint: str = app_entrypoint # plugin application initializing file
        self.app_dirpath: str = os.path.dirname(app_entrypoint) # plugin directory path
        self.name: str = os.path.basename(self.app_dirpath) # web url to serve application pages
        self.CONFIG_FILENAME = self.module
        self.available_models = {}
        self.routing_paths = []

    @property
    def request_paths(self) -> AppRequestPaths | None:
        """Request context for route resolution"""
        if not self.routing_paths and self.endpoint: return None
        return self.endpoint, self.routing_paths

    @property
    def backend_dirpath(self) -> str:
        """Directory path for site's python backend files (controllers, filters)"""
        return os.path.join(self.app_dirpath, self.BACKEND_DIRNAME)

    @property
    def contents_dirpath(self) -> str:
        """Directory path for site's theme folders"""
        return os.path.join(self.app_dirpath, self.CONTENTS_DIRNAME)

    @property
    def frontend_dirpath(self) -> str:
        """Directory path for site's theme folders"""
        return os.path.join(self.app_dirpath, self.FRONTEND_DIRNAME)

    @property
    def ssg_dirpath(self) -> str:
        """Directory path for site's static generated files"""
        return os.path.join(self.app_dirpath, self.SSG_DIRNAME)

    @property
    def app_ctx(self) -> AppCtx:
        return self._app_ctx or [self.name, self.endpoint, self.contents_dirpath, self.ssg_dirpath]

    def register_routing_dirpaths(self, dir_paths: list):
        """Registers a new pages directory path for resolving web based request"""
        for path in dir_paths:
            self.routing_paths.insert(0, path)
        pass

    def register_templates(self, dir_paths: list):
        """Registers additional paths for jinja templates. Templates will load in order of priority."""
        from jinja2 import FileSystemLoader

        if not hasattr(self.app.TemplateEnvironment, 'loader'): return None
        for path in dir_paths:
            if path in self.app.TemplateEnvironment.loader.loaders: continue
            self.app.TemplateEnvironment.loader.loaders.insert(0, FileSystemLoader(path))
        pass

    def insert(self, file_path: str, contents: dict) -> Parsely:
        """Creates a new file"""
        from pyonir import Parsely
        contents = Parsely.serializer(contents) if isinstance(contents, dict) else contents
        return Parsely.create_file(file_path, contents, app_ctx=self.app_ctx)

    @staticmethod
    def query_files(dir_path: str, app_ctx: tuple, model_type: any = None) -> list[Parsely]:
        from pyonir.utilities import process_contents
        # return PyonirCollection.query(dir_path, app_ctx, model_type)
        return process_contents(dir_path, app_ctx, model_type)

    @staticmethod
    def install_directory(plugin_src_directory: str, site_destination_directory: str):
        from pyonir.utilities import copy_assets
        copy_assets(plugin_src_directory, site_destination_directory)


class PyonirApp(PyonirBase):
    """Pyonir Application"""

    # Application data structures
    endpoint = '/'
    """Web url to access application resources."""

    server: PyonirServer = None
    """Starlette server instance"""

    TemplateEnvironment: TemplateEnvironment = None
    """Template environment for jinja templates"""

    plugins_installed: dict = dict()
    """Represents plugins installed within the site plugins directory"""

    plugins_activated: set = set()
    """All enabled plugins instances"""

    def __init__(self, app_entrypoint: str):
        from pyonir.utilities import generate_id, get_attr, process_contents
        from pyonir import __version__
        from pyonir.parser import Parsely
        self.SOFTWARE_VERSION = __version__
        self.get_attr = get_attr
        self.app_entrypoint: str = app_entrypoint # application main.py file or the initializing file
        self.app_dirpath: str = os.path.dirname(app_entrypoint) # application main.py file or the initializing file
        self.name: str = os.path.basename(self.app_dirpath) # web url to serve application pages
        self.SECRET_SAUCE = generate_id()
        self.SESSION_KEY = f"pyonir_{self.app_name}"
        self.configs = None
        self.routing_paths = [self.pages_dirpath, self.api_dirpath]
        Parsely._Filters['jinja'] = self.parse_jinja
        Parsely._Filters['pyformat'] = self.parse_pyformat

    @property
    def nginx_config_filepath(self):
        return os.path.join(self.app_dirpath, self.name + '.conf')

    @property
    def unix_socket_filepath(self):
        """WSGI socket file reference"""
        return os.path.join(self.app_dirpath, self.name+'.sock')

    @property
    def ssg_dirpath(self) -> str:
        """Directory path for site's static generated files"""
        return os.path.join(self.app_dirpath, self.SSG_DIRNAME)

    @property
    def logs_dirpath(self) -> str:
        """Directory path for site's log files"""
        return os.path.join(self.app_dirpath, 'logs')

    @property
    def backend_dirpath(self) -> str:
        """Directory path for site's python backend files (controllers, filters)"""
        return os.path.join(self.app_dirpath, self.BACKEND_DIRNAME)

    @property
    def contents_dirpath(self) -> str:
        """Directory path for site's theme folders"""
        return os.path.join(self.app_dirpath, self.CONTENTS_DIRNAME)

    @property
    def frontend_dirpath(self) -> str:
        """Directory path for site's theme folders"""
        return os.path.join(self.app_dirpath, self.FRONTEND_DIRNAME)

    @property
    def pages_dirpath(self) -> str:
        """Directory path to serve as file-based routing"""
        return os.path.join(self.contents_dirpath, self.PAGES_DIRNAME)

    @property
    def api_dirpath(self) -> str:
        """Directory path to serve API as file-based routing"""
        return os.path.join(self.contents_dirpath, self.API_DIRNAME)

    @property
    def plugins_dirpath(self) -> str:
        """Directory path to site's available plugins"""
        return os.path.join(self.app_dirpath, "plugins")

    @property
    def uploads_dirpath(self) -> str:
        """Directory path to site's available plugins"""
        return os.path.join(self.contents_dirpath, self.UPLOADS_DIRNAME)

    @property
    def jinja_filters_dirpath(self) -> str:
        """Directory path to site's available Jinja filters"""
        return os.path.join(self.backend_dirpath, "filters")

    @property
    def app_ctx(self) -> AppCtx:
        return self.name, self.endpoint, self.contents_dirpath, self.ssg_dirpath

    @property
    def env(self): return os.getenv('APPENV')

    @property
    def is_dev(self): return self.env == DEV_ENV

    @property
    def host(self): return self.get_attr(self.configs, 'app.host', '0.0.0.0') #if self.configs else '0.0.0.0'

    @property
    def port(self):
        return self.get_attr(self.configs, 'app.port', 5000) #if self.configs else 5000

    @property
    def protocol(self):return 'https' if self.is_secure else 'http'

    @property
    def is_secure(self):return self.get_attr(self.configs, 'app.use_ssl', False) #if self.configs else None

    @property
    def domain(self): return self.get_attr(self.configs, 'app.domain', self.host) # if self.configs else self.host

    def load_plugin(self, plugin: callable | list[callable]):
        """Make the plugin known to the pyonir application"""
        if isinstance(plugin, list):
            for plugin in plugin:
                self.load_plugin(plugin)
        else:
            plg_pkg_name = plugin.__module__.split('.').pop()
            self.plugins_installed[plg_pkg_name] = plugin

    # def _load_plugins(self):
    #     """Loads the plugin initializing object into runtime"""
    #     self.plugins_installed = load_modules_from(self.plugins_dirpath, only_packages=True)
    #     pass

    def _activate_plugins(self):
        """Active plugins enabled based on configurations"""
        is_configured = hasattr(self.configs, 'app') and hasattr(self.configs.app, 'enabled_plugins')
        for plg_id, plugin in self.plugins_installed.items():
            if is_configured and plg_id not in self.configs.app.enabled_plugins: continue
            self.plugins_activated.add(plugin(self))

    def install_sys_plugins(self):
        """Install pyonir plugins"""
        from pyonir.libs.plugins.navigation import Navigation
        self.plugins_installed['pyonir_navigation'] = Navigation

    def run_plugins(self, hook: PyonirHooks, data_value=None):
        if not hook or not self.plugins_activated: return
        hook = hook.name.lower()
        for plg in self.plugins_activated:
            if not hasattr(plg, hook): continue
            hook_method = getattr(plg, hook)
            hook_method(data_value, self)

    async def run_async_plugins(self, hook: PyonirHooks, data_value=None):
        if not hook or not self.plugins_activated: return
        hook_method_name = hook.lower()
        for plg in self.plugins_activated:
            if not hasattr(plg, hook_method_name): continue
            hook_method = getattr(plg, hook_method_name)
            await hook_method(data_value, self)

    def parse_jinja(self, string, context=None) -> str:
        """Render jinja template fragments"""
        if not context: context = {}
        if not self.TemplateEnvironment: return string
        try:
            return self.TemplateEnvironment.from_string(string).render(configs=self.configs, **context)
        except Exception as e:
            print(str(e))
            return string
            # raise

    def parse_pyformat(self, string, context=None) -> str:
        """Formats python template string"""
        ctx = self.TemplateEnvironment.globals if self.TemplateEnvironment else {}
        try:
            if context is not None: ctx.update(context)
            return string.format(**ctx)
        except Exception as e:
            print('parse_pyformat', e)
            return string

    def setup_templates(self):
        self.TemplateEnvironment = TemplateEnvironment(self)


    def setup_configs(self):
        """Setup site configurations and template environment"""
        from pyonir.utilities import process_contents, load_env
        self.configs = process_contents(os.path.join(self.contents_dirpath, self.CONFIGS_DIRNAME), self.app_ctx)
        envopts = load_env(os.path.join(self.app_dirpath, '.env'))
        setattr(self.configs, 'env', envopts)
        self.setup_templates()

    def run(self, routes: PyonirRouters, plugins: list[PyonirPlugin] =None):
        """Runs the Uvicorn webserver"""
        from .server import (setup_starlette_server, start_uvicorn_server,)

        # Initialize Server instance
        self.server = setup_starlette_server(self)
        # Initialize Application settings and templates
        self.setup_configs()
        self.install_sys_plugins()
        self._activate_plugins()

        # Run uvicorn server
        if self.SSG_IN_PROGRESS: return
        start_uvicorn_server(self, routes)


class TemplateEnvironment(Environment):

    def __init__(self, app: PyonirApp):
        from jinja2 import FileSystemLoader, ChoiceLoader
        from webassets import Environment as AssetsEnvironment
        from pyonir import PYONIR_JINJA_TEMPLATES_DIRPATH, PYONIR_JINJA_FILTERS_DIRPATH, PYONIR_JINJA_EXTS_DIRPATH
        from webassets.ext.jinja2 import AssetsExtension
        from pyonir.utilities import load_modules_from

        self.themes = PyonirThemes(os.path.join(app.frontend_dirpath, PyonirApp.THEMES_DIRNAME))

        jinja_template_paths = ChoiceLoader([FileSystemLoader(self.themes.active_theme.jinja_template_path),
                                             FileSystemLoader(PYONIR_JINJA_TEMPLATES_DIRPATH)])
        sys_filters = load_modules_from(PYONIR_JINJA_FILTERS_DIRPATH)
        app_filters = load_modules_from(app.jinja_filters_dirpath)
        installed_extensions = load_modules_from(PYONIR_JINJA_EXTS_DIRPATH, True)
        app_extensions = [AssetsExtension, *installed_extensions]
        app_filters = {**sys_filters, **app_filters}
        super().__init__(loader=jinja_template_paths, extensions=app_extensions)

        def url_for(path):
            rmaps = app.server.url_map if app.server else {}
            return rmaps.get(path, {}).get('path', app.ASSETS_ROUTE)

        app_active_theme = self.themes.active_theme
        #  Custom filters
        self.filters.update(**app_filters)
        # load assests tag
        self.assets_environment = AssetsEnvironment(app_active_theme.static_dirpath, app.ASSETS_ROUTE)
        # Add paths containing static assets
        # self.assets_environment.load_path.append(app_active_theme.static_dirpath)
        self.url_expire = False
        self.globals['url_for'] = url_for
        self.globals['configs'] = app.configs.app
        self.globals['request'] = None
        # self.globals.update(**app.jinja_template_globals)


    def add_jinja_path(self, path: str):
        pass

    def add_filter(self, filter: callable):
        name = filter.__name__
        print(name)
        self.filters.update({name: filter})
        pass


class PyonirThemes:
    """Represents sites available and active theme(s) within the frontend directory."""

    def __init__(self, theme_dirpath: str):
        self.themes_dirpath: str = theme_dirpath # directory path to available site themes
        self._available_themes: PyonirCollection | None = None # collection of themes available in frontend/themes directory

    @property
    def active_theme(self) -> Theme | None:
        from pyonir import Site
        from pyonir.parser import get_attr
        if not Site: return None
        self._available_themes = self._get_available_themes()
        site_theme = get_attr(Site.configs, 'app.theme_name')
        site_theme = self._available_themes.find(site_theme, from_attr='theme_dirname')
        return site_theme

    def _get_available_themes(self) -> PyonirCollection | None:
        from pyonir import Site
        if not Site: return None
        fe_ctx = list(Site.app_ctx)
        fe_ctx[2] = Site.frontend_dirpath
        pc = PyonirCollection.query(self.themes_dirpath, fe_ctx, include_only='README.md', data_model=Theme)
        return pc




import pyonir
from backend import router
from plugins.pyonir_forms import Forms
# Instantiate pyonir application
demo_app = pyonir.init(__file__)

# Install plugins from pyonir registry repo based on plugin identifier
# demo_app.install_plugins(['ADD_PYONIR_PLUGIN_IDENTIFIER'])
# Install custom plugins from local app
demo_app.load_plugin([Forms])

# Generate static website
# demo_app.generate_static_website()

# Run server
demo_app.run(routes=router)
